"""
This module is currently inactive.

It is reserved for future use and may contain experimental or planned features (dependence on refresh token).

Status:
    - Not in use (as of 2025-02-18)
    - Intended for future feature expansion

Module Name: get_order

This file contains the code for getting the status and details of an order.
"""
from typing import Any, Dict

from arklex.env.tools.tools import register_tool

# general GraphQL navigation utilities
from arklex.env.tools.shopify.utils_nav import *

# Customer API
from arklex.env.tools.shopify.utils_slots import ShopifySlots, ShopifyOutputs
from arklex.env.tools.shopify.utils import *
from arklex.env.tools.shopify.auth_utils import *

description = "Get the status and details of an order."
slots = [
    ShopifySlots.REFRESH_TOKEN,
    ShopifySlots.ORDER_ID,
    *PAGEINFO_SLOTS
]
outputs = [
    ShopifyOutputs.ORDERS_DETAILS,
    *PAGEINFO_OUTPUTS
]

ORDERS_NOT_FOUND = "error: order not found"
errors = [ORDERS_NOT_FOUND]

@register_tool(description, slots, outputs, lambda x: x not in errors)
def get_order(refresh_token, order_id: str, **kwargs) -> str:
    nav = cursorify(kwargs)
    if not nav[1]: 
        return nav[0]
    
    try:
        body = f'''
            query {{ 
                order (id: "{order_id}") {{ 
                    id
                    name
                    totalPrice {{
                        amount
                    }}
                    lineItems({nav[0]}) {{
                        nodes {{
                            id
                            name
                            quantity
                        }}
                        pageInfo {{
                            endCursor
                            hasNextPage
                            hasPreviousPage
                            startCursor
                        }}
                    }}
                }} 
            }}
        '''
        try:
            auth = {'Authorization': get_access_token(refresh_token)}
        except:
            return AUTH_ERROR
        
        try:
            response = make_query(customer_url, body, {}, customer_headers | auth)['data']['order']
        except Exception as e:
            return f"error: {e}"
        
        pageInfo = response['lineItems']['pageInfo']
        return response, pageInfo
    except Exception as e:
        return ORDERS_NOT_FOUND



import json
import random
from arklex.evaluation.extract_conversation_info import build_intent_graph
from arklex.evaluation.chatgpt_utils import chatgpt_chatbot, query_chatbot, flip_hist, filter_convo, generate_goals
from arklex.evaluation.get_documents import load_docs

def sampling_paths(start_node, graph, path_length, max_turns, intents):
    children = list(graph.successors(start_node))
    if path_length >= max_turns or len(children) == 0:
        return intents
    weights = []
    for c in children:
        weights.append(graph.get_edge_data(start_node, c)['weight'])
    next_node = random.choices(children, weights)[0]
    intents.append(next_node)
    return sampling_paths(next_node, graph, path_length+1, max_turns, intents)

def get_paths(G, num_paths, max_turns):
    my_paths = []
    for i in range(num_paths):
        my_path = sampling_paths('start', G, 0, max_turns, ['start'])
        my_paths.append(my_path[1:])
    return my_paths

def interact(intent_path, summary, model_api, model_params):
    history = []
    instructional_prompt = 'Replicate the behavior of a human customer. You are interacting with customer service chatbot for the following company: ' + summary
    start_text = "Begin the conversation as a human customer with the following intent: " + intent_path[0]
    history.append({'role': 'system','content': instructional_prompt})
    history.append({'role': 'user', 'content': start_text})
    for i in range(len(intent_path)):
        intent = intent_path[i]
        output = chatgpt_chatbot(history) 
        history.append({'role': 'assistant', 'content': output, 'intent': intent})
        response_data = query_chatbot(model_api, filter_convo(history), model_params)
        answer = response_data["answer"]
        answer = answer.replace('\n', ' ')
        model_params = response_data.get("parameters", model_params)
        if i < len(intent_path) - 1:
            intent = intent_path[i+1]
        history.append({'role': 'user', 'content': answer + '\nRespond to this utterance with the following intent: ' + intent + '\nMake sure your response is natural and follows the flow of the conversation. For example, if the bot asks you a question make sure you answer it.'})
    return history

def generate_labeled_convos(intent_paths, summary, model_api, model_params):
    convos = []
    model_params = {}
    for intent_path in intent_paths:
        convo = interact(intent_path, summary, model_api, model_params)
        convos.append(flip_hist(filter_convo(convo)))
    return convos

def get_labeled_convos(first_pass_data, model_api, synthetic_data_params, model_params, config):
    intent_graph = build_intent_graph(first_pass_data)
    intent_paths = get_paths(intent_graph, synthetic_data_params['num_convos'], synthetic_data_params['max_turns'])
    summary = config['intro']
    convos = generate_labeled_convos(intent_paths, summary, model_api, model_params)
    return convos


if __name__ == "__main__":
    with open('temp_files/p1_sample_convos_labeled.json') as f:
        data = json.load(f)
    
    with open('temp_files/richtech_config.json') as f:
        config = json.load(f)

    model_api = "http://adaptation.cs.columbia.edu:55231/qa/richtech/v1alpha1"
    synthetic_data_params = {'num_convos': 2, 'num_goals': 3, 'max_turns': 10}
    model_params = {}

    labeled_convos = get_labeled_convos(data, model_api, synthetic_data_params, model_params, config)

    with open('files/p2_sample_convos.json', 'w') as f:
        json.dump(labeled_convos, f, indent=5)


import os
import sqlite3
import logging
from langchain_openai import ChatOpenAI

from arklex.utils.model_config import MODEL

DBNAME = 'show_booking_db.sqlite'
USER_ID = "user_be6e1836-8fe9-4938-b2d0-48f810648e72"

logger = logging.getLogger(__name__)

SLOTS = {
    "show_name": {
        "name": "show_name",
        "type": "string",
        "value": "",
        "description": "Name of the show",
        "prompt": "Please provide the name of the show"
    },
    "location": {
        "name": "location",
        "type": "string",
        "value": "",
        "description": "Location of the show",
        "prompt": "Please provide the location of the show"
    },
    "date": {
        "name": "date",
        "type": "date",
        "value": "",
        "description": "Date of the show",
        "prompt": "Please provide the date of the show"
    },
    "time": {
        "name": "time",
        "type": "time",
        "value": "",
        "description": "Time of the show",
        "prompt": "Please provide the time of the show"
    }
}

LOG_IN_FAILURE = "Failed to login. Please ensure user ID and database information is correct"
NO_SHOW_MESSAGE = "Show is not found. Please check whether the information is correct."
MULTIPLE_SHOWS_MESSAGE = "There are multiple shows found. Please provide more details."
NO_BOOKING_MESSAGE = "You have not booked any show."

class Booking:
    db_path = None
    llm = ChatOpenAI(model=MODEL["model_type_or_path"], timeout=30000)
    user_id = USER_ID
    # actions = {
    #     "SearchShow": "Search for shows", 
    #     "BookShow": "Book a show", 
    #     "CheckBooking": "Check details of booked show(s)",
    #     "CancelBooking": "Cancel a booking",
    #     "Others": "Other actions not mentioned above"
    # }

booking = Booking()

def log_in():
    booking.db_path = os.path.join(os.environ.get("DATA_DIR"), DBNAME)
    conn = sqlite3.connect(booking.db_path)
    cursor = conn.cursor()
    cursor.execute("SELECT 1 FROM user WHERE id = ?", (booking.user_id,))
    result = cursor.fetchone()
    if result is None:
        logger.info(f"User {booking.user_id} not found in the database.")
    else:
        logger.info(f"User {booking.user_id} successfully logged in.")
    return result is not None


import os
import sys
root_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.append(root_dir)
import json
import argparse
import logging
import subprocess
import subprocess
import uuid
import atexit
from dotenv import load_dotenv


from langchain_openai import ChatOpenAI

from arklex.orchestrator.orchestrator import AgentOrg
from arklex.utils.utils import init_logger
from arklex.utils.model_config import MODEL
from arklex.orchestrator.generator.generator import Generator
from arklex.env.env import DefaulResourceInitializer
from arklex.env.tools.tools import Tool

from benchmark.tau_bench.envs.retail.tools import ALL_TOOLS
from benchmark.tau_bench.envs import get_env
from benchmark.tau_bench.types import RunConfig
from benchmark.tau_bench.run import run
from benchmark.tau_bench.envs.retail.data import load_data

load_dotenv()

API_PORT = "55135"
NLUAPI_ADDR = f"http://localhost:{API_PORT}/nlu"
SLOTFILLAPI_ADDR = f"http://localhost:{API_PORT}/slotfill"

tool_name_class_map = {}

def get_tool_name_class_map():
    tool_map = {}
    for tool in ALL_TOOLS:
        name = tool.get_info()["function"]["name"]
        tool_map[name] = tool
    return tool_map

class TauBenchResourceInitializer(DefaulResourceInitializer):
    @staticmethod
    def init_tools(tools):
        tool_name_class_map = get_tool_name_class_map()
        tool_registry = {}
        def tool_lambda(val): return lambda: val
        for tool_id, tool_info in tools.items():
            tool_name = tool_info["name"]
            tool_original_class = tool_name_class_map[tool_name]
            tool_func = tool_original_class.invoke
            tool_key = tool_name
            tool_desc = tool_info["description"]
            params = tool_original_class.get_info()["function"]["parameters"]
            tool_slots = []
   
            for param_name, param_info in params["properties"].items():
                slot = {}
                slot["name"] = param_name
                slot["type"] = param_info["type"]
                slot["items"] = param_info.get("items", {})
                slot["description"] = param_info["description"]
                prompt_param_name = param_name.replace("_", " ")
                slot["prompt"] = f"In order to proceed, please provide the {prompt_param_name}"
                slot["required"] = param_name in params["required"]
                tool_slots.append(slot)
            tool_output = []
            isComplete = lambda x: True
            
            tool = tool_lambda(Tool(tool_func, tool_key, tool_desc, tool_slots, tool_output, isComplete))

            tool_registry[tool_id] = {
                "name": tool_name,
                "description": tool_desc,
                "execute": tool,
                "fixed_args": {"data": load_data()},
            }
        return tool_registry

def generate_tau_bench_config(output_dir):
    retain_tools = ALL_TOOLS
    tools = {}
    for tool in retain_tools:
        tool_id = str(uuid.uuid1())
        tools[tool_id] = {}
        tools[tool_id]["name"] = tool.get_info()["function"]["name"]
        tools[tool_id]["description"] = tool.get_info()["function"]["description"]
    retail_config = {
        "role": "Retail Agent",
        "user_objective": "The core goal of the agent is to assist a single, authenticated user per conversation in managing their retail orders—resolving any questions, cancellations, modifications, exchanges, or returns—while strictly following the rules and confirmation steps set by the retail policy.",
        "builder_objective": "Users want a convenient, reliable way to manage their orders—whether that means updating their shipping address, switching payment methods, or returning/exchanging items they've received. They come to the Retail Agent because they need to quickly resolve questions about their orders, get real-time updates on shipping statuses, and handle any necessary cancellations or modifications with confidence that every action is confirmed and secure.",
        "domain": "retail",
        "intro": "Welcome to the Retail Agent service. By confirming your identity, I can help you with detailed information on your orders, profile, and products. If you need to cancel or modify any pending orders, change your shipping address, payment method, or exchange/return delivered items, I can guide you through it step by step. I will always ask you to confirm before making any changes to ensure accuracy and security.",
        "task_docs": [{
            "source": "https://raw.githubusercontent.com/sierra-research/tau-bench/refs/heads/main/tau_bench/envs/retail/wiki.md",
            "num": 20
        }],
        "rag_docs": [],
        "tasks": [],
        "workers": [
            {"id": "26bb6634-3bee-417d-ad75-23269ac17bc3", "name": "MessageWorker", "path": "message_worker.py"},
        ],
        "tools": tools,
        "tool_initialization": False
    }
    with open(os.path.join(output_dir, 'config.json'), 'w') as f:
        json.dump(retail_config, f, indent=4)

def generate_taskgraph(config_file, output_dir):
    model = ChatOpenAI(model=MODEL["model_type_or_path"], timeout=30000)
    resource_initializer = TauBenchResourceInitializer()
    generator = Generator(args, config_file, model, output_dir, resource_initializer)
    taskgraph_filepath = generator.generate()
    # Update the task graph with the API URLs
    task_graph = json.load(open(os.path.join(root_dir, taskgraph_filepath)))
    task_graph["nluapi"] = NLUAPI_ADDR
    task_graph["slotfillapi"] = SLOTFILLAPI_ADDR
    with open(taskgraph_filepath, "w") as f:
        json.dump(task_graph, f, indent=4)



nlu_process = None
def terminate_subprocess():
    """Terminate the FastAPI subprocess."""
    global nlu_process
    if nlu_process and nlu_process.poll() is None:  # Check if process is running
        logger.info(f"Terminating FastAPI process with PID: {nlu_process.pid}")
        nlu_process.terminate()  # Send SIGTERM
        nlu_process.wait()  # Ensure it stops
        logger.info("FastAPI process terminated.")


atexit.register(terminate_subprocess)

def start_apis():
    """Start the FastAPI subprocess and update task graph API URLs."""
    global nlu_process
    command = [
        "uvicorn",
        "arklex.orchestrator.NLU.api:app",  # Replace with proper import path
        "--port", API_PORT,
        "--host", "0.0.0.0",
        "--log-level", "warning"
    ]

    # Redirect FastAPI logs to a file
    with open(os.path.join(root_dir, "logs", "api.log"), "w") as log_file:
        nlu_process = subprocess.Popen(
            command,
            stdout=log_file,  # Redirect stdout to a log file
            stderr=subprocess.STDOUT,  # Redirect stderr to the same file
            start_new_session=True  # Run in a separate process group
        )
    logger.info(f"Started FastAPI process with PID: {nlu_process.pid}")

def run_tau_bench_eval(
        taskgraph_dir,
        output_dir,
        num_trials,
        task_ids,
        env,
        task_split="test",
        user_strategy="llm",
        max_concurrency=10,
):
 
    start_index = 0
    end_index = -1
    seed=10
    shuffle=0
    
    
    config = RunConfig(
        user_model_provider="openai",
        user_model="gpt-4o",
        num_trials=num_trials,
        env=env,
        task_split=task_split,
        start_index=start_index,
        end_index=end_index,
        task_ids=task_ids,
        output_dir=output_dir,
        max_concurrency=max_concurrency,
        seed=seed,
        shuffle=shuffle,
        user_strategy=user_strategy,
        taskgraph_dir=taskgraph_dir
    )
    run(config)
    



if __name__ == "__main__":
    '''
        Provide --output-dir
    '''
    parser = argparse.ArgumentParser()
    parser.add_argument('--output-dir', type=str, default="./examples/tau_bench")
    parser.add_argument('--num-trials', type=int, default=1)
    parser.add_argument('--env', type=str, default="retail", choices=["retail"])
    parser.add_argument('--task-ids', type=list, default=None)

    parser.add_argument('--model_api', type=str, default="http://127.0.0.1:8000/eval/chat")
    parser.add_argument('--model', type=str, default=MODEL["model_type_or_path"])
    parser.add_argument('--log-level', type=str, default="WARNING", choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"])
    args = parser.parse_args()
    
    
    assert args.output_dir is not None, "Output dir must be provided"

    os.makedirs(os.path.join(args.output_dir, 'eval'), exist_ok=True)
    os.makedirs(os.path.join(args.output_dir, 'temp'), exist_ok=True)

    temp_output_dir = os.path.join(args.output_dir, 'temp')
    eval_output_dir = os.path.join(args.output_dir, 'eval')

    MODEL["model_type_or_path"] = args.model
    log_level = getattr(logging, args.log_level.upper(), logging.INFO)
    logger = init_logger(log_level=log_level, filename=os.path.join(root_dir, "logs", "tau_bench_eval.log"))
    
    # generate_tau_bench_config(temp_output_dir)
    # config_file = os.path.join(temp_output_dir, 'config.json')
    # generate_taskgraph(config_file, temp_output_dir)

    start_apis()
    run_tau_bench_eval(
        taskgraph_dir=temp_output_dir,
        output_dir=eval_output_dir,
        num_trials=args.num_trials,
        env=args.env,
        task_ids=args.task_ids
    )
    


import logging
import time
from pathlib import Path
from typing import List
import requests
import pickle
import uuid
import argparse
import os

from webdriver_manager.chrome import ChromeDriverManager
from selenium import webdriver
from bs4 import BeautifulSoup
from urllib.parse import urljoin
import networkx as nx
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain_core.documents import Document


# Configure logging
logging.basicConfig(
    level=logging.INFO,  
    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
    handlers=[logging.StreamHandler()]
)
logger = logging.getLogger(__name__)

CHROME_DRIVER_VERSION = "125.0.6422.7"

class URLObject:
    def __init__(self, id: str, url: str):
        self.id = id
        self.url = url


class CrawledURLObject(URLObject):
    def __init__(
        self,
        id: str,
        url: str,
        content: str,
        metadata={},
        is_chunk=False,
        is_error=False,
        error_message=None,
    ):
        super().__init__(id, url)
        self.content = content
        self.metadata = metadata
        self.is_chunk = is_chunk
        self.is_error = is_error
        self.error_message = error_message


class Loader:
    def __init__(self):
        pass

    def to_crawled_obj(self, url_list: List[str]):    
        url_objs = [URLObject(str(uuid.uuid4()), url) for url in url_list]
        crawled_url_objs = self.crawl_urls(url_objs)
        return crawled_url_objs

    def crawl_urls(self, url_objects: list[URLObject]) -> List[CrawledURLObject]:
        logger.info(f"Start crawling {len(url_objects)} urls")
        options = webdriver.ChromeOptions()
        options.add_argument("--no-sandbox")
        options.add_argument("--headless")
        options.add_argument("--disable-dev-shm-usage")
        options.add_argument("--disable-gpu")
        options.add_argument("--disable-extensions")
        options.add_argument("--disable-infobars")
        options.add_argument("--remote-debugging-pipe")
        chrome_driver_path = Path(ChromeDriverManager(driver_version=CHROME_DRIVER_VERSION).install())
        options.binary_location = str(chrome_driver_path.parent.absolute())
        logger.info(f"chrome binary location: {options.binary_location}")
        driver = webdriver.Chrome(options=options)

        docs: List[CrawledURLObject] = []
        for url_obj in url_objects:
            try:
                logger.info(f"loading url: {url_obj.url}")
                driver.get(url_obj.url)
                time.sleep(2)
                html = driver.page_source
                soup = BeautifulSoup(html, "html.parser")

                text_list = []
                for string in soup.strings:        
                    if string.find_parent("a"):
                        href = urljoin(url_obj.url, string.find_parent("a").get("href"))
                        if href.startswith(url_obj.url):
                            text = f"{string} {href}"
                            text_list.append(text)
                    elif string.strip():
                        text_list.append(string)
                text_output = "\n".join(text_list)
                
                title = url_obj.url
                for title in soup.find_all("title"):
                    title = title.get_text()
                    break

                docs.append(
                    CrawledURLObject(
                        id=url_obj.id,
                        url=url_obj.url,
                        content=text_output,
                        metadata={"title": title, "source": url_obj.url},
                    )
                )

            except Exception as err:
                logger.info(f"error crawling {url_obj}")
                logger.error(err)
                docs.append(
                    CrawledURLObject(
                        id=url_obj.id,
                        url=url_obj.url,
                        content=None,
                        metadata={"title": url_obj.url, "source": url_obj.url},
                        is_error=True,
                        error_message=str(err),
                    )
                )
        driver.quit()
        return docs

    def get_all_urls(self, base_url: str, max_num: int) -> List[str]:
        logger.info(f"Getting all pages for base url: {base_url}, maximum number is: {max_num}")
        urls_visited = []
        base_url = base_url.split("#")[0].rstrip("/")
        urls_to_visit = [base_url]
        
        while urls_to_visit:
            if len(urls_visited) >= max_num:
                break
            current_url = urls_to_visit.pop(0)
            if current_url not in urls_visited:
                urls_visited.append(current_url)
                new_urls = self.get_outsource_urls(current_url, base_url)
                urls_to_visit.extend(new_urls)
                urls_to_visit = list(set(urls_to_visit))
        logger.info(f"URLs visited: {urls_visited}")
        return sorted(urls_visited[:max_num])
    
    def get_outsource_urls(self, curr_url: str, base_url: str):
        headers = {
            'User-Agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/14.0.3 Safari/605.1.15'
        } 
        new_urls = list()
        try:
            response = requests.get(curr_url, headers=headers, timeout=10)
            # Check if the request was successful
            if response.status_code == 200:
                soup = BeautifulSoup(response.text, 'html.parser')
                for link in soup.find_all("a"):
                    try:
                        full_url = urljoin(curr_url, link.get("href"))
                        full_url = full_url.split("#")[0].rstrip("/")
                        if self._check_url(full_url, base_url):
                            new_urls.append(full_url)
                    except Exception as err:
                        logger.error(f"Fail to process sub-url {link.get('href')}: {err}")
            else:
                logger.error(f"Failed to retrieve page {curr_url}, status code: {response.status_code}")
        except Exception as err:
            logger.error(f"Fail to get the page from {curr_url}: {err}")
        return list(set(new_urls))
    
    def _check_url(self, full_url, base_url):
        kw_list = ['.pdf', '.jpg', '.png', '.docx', '.xlsx', '.pptx', '.zip', ".jpeg"]
        if full_url.startswith(base_url) and full_url and not any(kw in full_url for kw in kw_list) and full_url != base_url:
            return True
        return False

    def get_candidates_websites(self, urls: List[CrawledURLObject], top_k: int) -> List[str]:
        """Based on the pagerank algorithm of the crawled websites, return the top 10 websites. 
        The reason why we can do that is because we have the hreqs of the including <a> tags in the content of the website.
        So we can use that to construct the edges and then use the tool from networkx to get the pagerank of the websites.
        """
        
        nodes = []
        edges = []
        url_to_id_mapping = {}
        for url in urls:
            url_to_id_mapping[url.url] = url.id

        for url in urls:
            if url.is_error:
                continue
            for url_key in url_to_id_mapping:
                if url_key in url.content:
                    edge = [url.id, url_to_id_mapping[url_key]]
                    edges.append(edge)

            node = [url.id, {"url": url.url, "content": url.content, "metadata": url.metadata}]
            nodes.append(node)
        
        self.graph = nx.DiGraph(name="website graph")
        self.graph.add_nodes_from(nodes)
        self.graph.add_edges_from(edges)
        pr = nx.pagerank(self.graph, alpha=0.9)
        # sort the pagerank values in descending order
        sorted_pr = sorted(pr.items(), key=lambda x: x[1], reverse=True)
        logger.info(f"pagerank results: {sorted_pr}")
        # get the top websites
        top_k_websites = sorted_pr[:top_k]
        urls_candidates = [self.graph.nodes[url_id] for url_id, _ in top_k_websites]
        urls_cleaned = [doc for doc in urls_candidates if doc]
        return urls_cleaned
    
    @staticmethod
    def save(file_path: str, docs: List[CrawledURLObject]):
        with open(file_path, "wb") as f:
            pickle.dump(docs, f)
    
    @classmethod
    def chunk(cls, url_objs: List[CrawledURLObject]) -> List[CrawledURLObject]:
        text_splitter = RecursiveCharacterTextSplitter.from_tiktoken_encoder(encoding_name="cl100k_base", chunk_size=200, chunk_overlap=40)
        docs = []
        langchain_docs = []
        for url_obj in url_objs:
            if url_obj.is_error or url_obj.content is None:
                logger.info(f"Skip url: {url_obj.url} because of error or no content")
                continue
            elif url_obj.is_chunk:
                logger.info(f"Skip url: {url_obj.url} because it has been chunked")
                docs.append(url_obj)
                continue
            splitted_text = text_splitter.split_text(url_obj.content)
            for i, txt in enumerate(splitted_text):
                doc = CrawledURLObject(
                    id=url_obj.id+"_"+str(i),
                    url=url_obj.url,
                    content=txt,
                    metadata=url_obj.metadata,
                    is_chunk=True,
                )
                docs.append(doc)
                langchain_docs.append(Document(page_content=txt, metadata={"source": url_obj.url}))
        return langchain_docs




import logging
from typing import Any, Iterator, Union

from langgraph.graph import StateGraph, START
from langchain.prompts import PromptTemplate
from langchain_openai import ChatOpenAI
from langchain_core.output_parsers import StrOutputParser

from arklex.env.workers.worker import BaseWorker, register_worker
from arklex.env.prompts import load_prompts
from arklex.types import EventType
from arklex.utils.utils import chunk_string
from arklex.utils.graph_state import MessageState
from arklex.utils.model_config import MODEL
from arklex.utils.model_provider_config import PROVIDER_MAP


logger = logging.getLogger(__name__)


@register_worker
class MessageWorker(BaseWorker):

    description = "The worker that used to deliver the message to the user, either a question or provide some information."

    def __init__(self):
        super().__init__()
        self.llm = PROVIDER_MAP.get(MODEL['llm_provider'], ChatOpenAI)(
            model=MODEL["model_type_or_path"], timeout=30000
        )
        self.action_graph = self._create_action_graph()

    def generator(self, state: MessageState) -> MessageState:
        # get the input message
        user_message = state['user_message']
        orchestrator_message = state['orchestrator_message']
        message_flow = state.get('response', "") + "\n" + state.get("message_flow", "")

        # get the orchestrator message content
        orch_msg_content = "None" if not orchestrator_message.message else orchestrator_message.message
        orch_msg_attr = orchestrator_message.attribute
        direct_response = orch_msg_attr.get('direct_response', False)
        if direct_response:
            state["message_flow"] = ""
            state["response"] = orch_msg_content
            return state
        
        prompts = load_prompts(state["bot_config"])
        if message_flow and message_flow != "\n":
            prompt = PromptTemplate.from_template(prompts["message_flow_generator_prompt"])
            input_prompt = prompt.invoke({"sys_instruct": state["sys_instruct"], "message": orch_msg_content, "formatted_chat": user_message.history, "context": message_flow})
        else:
            prompt = PromptTemplate.from_template(prompts["message_generator_prompt"])
            input_prompt = prompt.invoke({"sys_instruct": state["sys_instruct"], "message": orch_msg_content, "formatted_chat": user_message.history})
        logger.info(f"Prompt: {input_prompt.text}")
        chunked_prompt = chunk_string(input_prompt.text, tokenizer=MODEL["tokenizer"], max_length=MODEL["context"])
        final_chain = self.llm | StrOutputParser()
        answer = final_chain.invoke(chunked_prompt)

        state["message_flow"] = ""
        state["response"] = answer
        return state
    
    def choose_generator(self, state: MessageState):
        if state["is_stream"]:
            return "stream_generator"
        return "generator"
    
    def stream_generator(self, state: MessageState) -> MessageState:
        # get the input message
        user_message = state['user_message']
        orchestrator_message = state['orchestrator_message']
        message_flow = state.get('response', "") + "\n" + state.get("message_flow", "")

        # get the orchestrator message content
        orch_msg_content = "None" if not orchestrator_message.message else orchestrator_message.message
        orch_msg_attr = orchestrator_message.attribute
        direct_response = orch_msg_attr.get('direct_response', False)
        if direct_response:
            state["message_flow"] = ""
            state["response"] = orch_msg_content
            return state
        
        prompts = load_prompts(state["bot_config"])
        if message_flow and message_flow != "\n":
            prompt = PromptTemplate.from_template(prompts["message_flow_generator_prompt"])
            input_prompt = prompt.invoke({"sys_instruct": state["sys_instruct"], "message": orch_msg_content, "formatted_chat": user_message.history, "context": message_flow})
        else:
            prompt = PromptTemplate.from_template(prompts["message_generator_prompt"])
            input_prompt = prompt.invoke({"sys_instruct": state["sys_instruct"], "message": orch_msg_content, "formatted_chat": user_message.history})
        logger.info(f"Prompt: {input_prompt.text}")
        chunked_prompt = chunk_string(input_prompt.text, tokenizer=MODEL["tokenizer"], max_length=MODEL["context"])
        final_chain = self.llm | StrOutputParser()
        answer = ""
        for chunk in final_chain.stream(chunked_prompt):
            answer += chunk
            state["message_queue"].put({"event": EventType.CHUNK.value, "message_chunk": chunk})

        state["message_flow"] = ""
        state["response"] = answer
        return state

    def _create_action_graph(self):
        workflow = StateGraph(MessageState)
        # Add nodes for each worker
        workflow.add_node("generator", self.generator)
        workflow.add_node("stream_generator", self.stream_generator)
        # Add edges
        # workflow.add_edge(START, "generator")
        workflow.add_conditional_edges(START, self.choose_generator)
        return workflow

    def execute(self, msg_state: MessageState):
        graph = self.action_graph.compile()
        result = graph.invoke(msg_state)
        return result




# Copyright Sierra

import random
from hashlib import sha256
from benchmark.tau_bench.envs.tool import Tool
from typing import Any, Callable, Dict, List, Type, Optional, Set, Union, Tuple

from benchmark.tau_bench.envs.user import load_user, UserStrategy
from benchmark.tau_bench.types import (
    Action,
    Task,
    EnvInfo,
    EnvResetResponse,
    EnvResponse,
    RewardResult,
    RewardOutputInfo,
    RewardActionInfo,
    RESPOND_ACTION_NAME,
)

ToHashable = Union[
    str, int, float, Dict[str, "ToHashable"], List["ToHashable"], Set["ToHashable"]
]
Hashable = Union[str, int, float, Tuple["Hashable"], Tuple[Tuple[str, "Hashable"]]]


def to_hashable(item: ToHashable) -> Hashable:
    if isinstance(item, dict):
        return tuple((key, to_hashable(value)) for key, value in sorted(item.items()))
    elif isinstance(item, list):
        return tuple(to_hashable(element) for element in item)
    elif isinstance(item, set):
        return tuple(sorted(to_hashable(element) for element in item))
    else:
        return item


def consistent_hash(
    value: Hashable,
) -> str:
    return sha256(str(value).encode("utf-8")).hexdigest()


class Env(object):
    def __init__(
        self,
        data_load_func: Callable[[], Dict[str, Any]],
        tools: List[Type[Tool]],
        tasks: List[Task],
        wiki: str,
        rules: List[str],
        user_strategy: Union[str, UserStrategy],
        user_model: str,
        user_provider: Optional[str] = None,
        task_index: Optional[int] = None,
    ) -> None:
        super().__init__()
        self.data_load_func = data_load_func
        self.data = data_load_func()
        self.tools_map: Dict[str, Type[Tool]] = {
            tool.get_info()["function"]["name"]: tool for tool in tools
        }
        self.tools_info = [tool.get_info() for tool in tools]
        self.terminate_tools = []
        self.tasks = tasks
        if task_index is not None:
            self.task_index = task_index
        else:
            self.task_index = random.randint(0, len(tasks))
        self.task = tasks[self.task_index]
        self.wiki = wiki
        self.rules = rules
        self.user = load_user(
            user_strategy=user_strategy, model=user_model, provider=user_provider
        )
        self.actions: List[Action] = []

    def reset(self, task_index: Optional[int] = None) -> EnvResetResponse:
        if task_index is None:
            task_index = random.randint(0, len(self.tasks))
        self.task_index = task_index
        self.data = self.data_load_func()
        self.task = self.tasks[task_index]
        self.actions = []
        initial_observation = self.user.reset(instruction=self.task.instruction)
        return EnvResetResponse(
            observation=initial_observation, info=EnvInfo(task=self.task, source="user")
        )

    def step(self, action: Action) -> EnvResponse:
        self.actions.append(action)

        info = EnvInfo(task=self.task)
        reward = 0
        done = False
        if action.name == RESPOND_ACTION_NAME:
            observation = self.user.step(action.kwargs["content"])
            info.source = "user"
            done = "###STOP###" in observation
        elif action.name in self.tools_map:
            try:
                observation = self.tools_map[action.name].invoke(
                    data=self.data, **action.kwargs
                )
            except Exception as e:
                observation = f"Error: {e}"
            info.source = action.name
            if action.name in self.terminate_tools:
                done = True
        else:
            observation = f"Unknown action {action.name}"
            info.source = action.name

        if done:
            reward_res = self.calculate_reward()
            reward = reward_res.reward
            info.reward_info = reward_res
            info.user_cost = self.user.get_total_cost()
        return EnvResponse(observation=observation, reward=reward, done=done, info=info)

    def get_data_hash(self) -> str:
        return consistent_hash(to_hashable(self.data))

    def calculate_reward(self) -> RewardResult:
        data_hash = self.get_data_hash()
        reward = 1.0
        actions = [
            action for action in self.task.actions if action.name != RESPOND_ACTION_NAME
        ]

        # Check if the database changes are correct. If they are not correct, then we set the reward to 0.
        # TODO: cache gt_data_hash in tasks.py (low priority)
        self.data = self.data_load_func()
        for action in self.task.actions:
            if action.name not in self.terminate_tools:
                self.step(action)
        gt_data_hash = self.get_data_hash()
        info = RewardActionInfo(
            r_actions=data_hash == gt_data_hash, gt_data_hash=gt_data_hash
        )
        if not info.r_actions:
            reward = 0.0

        if len(self.task.outputs) > 0:
            # check outputs
            r_outputs = 1.0
            outputs = {}
            for output in self.task.outputs:
                found = False
                for action in self.actions:
                    if (
                        action.name == RESPOND_ACTION_NAME
                        and output.lower()
                        in action.kwargs["content"].lower().replace(",", "")
                    ):
                        found = True
                        break
                outputs[output] = found
                if not found:
                    r_outputs = 0.0
                    reward = 0.0
            info = RewardOutputInfo(r_outputs=r_outputs, outputs=outputs)
            
        return RewardResult(reward=reward, info=info, actions=actions)



"""
This module is currently inactive.

It is reserved for future use and may contain experimental or planned features.

Status:
    - Not in use (as of 2025-02-18)
    - Intended for future feature expansion

Module Name: auth

This file implements the function of Shopify authentication.
"""
# from arklex.tools.shopify_new.auth_utils import get_auth_link, get_refresh_token, get_access_token
# from arklex.tools.shopify_new.auth_server import authenticate_server
from auth_utils import get_auth_link, get_refresh_token, get_access_token
from auth_server import authenticate_server

import os

def authenticate():
    auth_link = get_auth_link()
    print("Authenticate Link here: ", auth_link)
    code = authenticate_server()
    refresh_token = get_refresh_token(code)
    # access_token = get_access_token(refresh_token)
    return refresh_token

if __name__ == "__main__":
    refresh_token = authenticate()
    
    os.environ["SHOPIFY_CUSTOMER_API_REFRESH_TOKEN"] = refresh_token
    print(f"Refresh token: {refresh_token}")


import json
from typing import Any, Dict
import shopify
import logging

# general GraphQL navigation utilities
from arklex.env.tools.shopify.utils_slots import ShopifySlots, ShopifyOutputs
from arklex.env.tools.shopify.utils_nav import *
from arklex.env.tools.shopify.utils import authorify_admin

# Admin API
from arklex.env.tools.tools import register_tool

from arklex.utils.model_provider_config import PROVIDER_MAP
from arklex.utils.model_config import MODEL
from langchain_openai import ChatOpenAI


logger = logging.getLogger(__name__)


description = "Search products by string query. If no products are found, the function will return an error message."
slots = [
    ShopifySlots.SEARCH_PRODUCT_QUERY,
    *PAGEINFO_SLOTS
] 
outputs = [
    ShopifyOutputs.PRODUCT_ID,
    *PAGEINFO_OUTPUTS
]
PRODUCT_SEARCH_ERROR = "error: product search failed"
NO_PRODUCTS_FOUND_ERROR = "no products found"

errors = [
    PRODUCT_SEARCH_ERROR,
    NO_PRODUCTS_FOUND_ERROR,
    NAVIGATE_WITH_NO_CURSOR,
    NO_NEXT_PAGE,
    NO_PREV_PAGE
]

@register_tool(description, slots, outputs, lambda x: x[0] not in errors, True)
def search_products(product_query: str, **kwargs) -> str:
    nav = cursorify(kwargs)
    if not nav[1]:
        return nav[0]
    auth = authorify_admin(kwargs)
    if auth["error"]:
        return auth["error"]
    
    try:
        with shopify.Session.temp(**auth["value"]):
            response = shopify.GraphQL().execute(f"""
                {{
                    products ({nav[0]}, query: "{product_query}") {{
                        nodes {{
                            id
                            title
                            description
                            onlineStoreUrl
                            images(first: 1) {{
                                edges {{
                                    node {{
                                        src
                                        altText
                                    }}
                                }}
                            }}
                            variants (first: 3) {{
                                nodes {{
                                    displayName
                                    id
                                    price
                                    inventoryQuantity
                                }}
                            }}
                        }}
                        pageInfo {{
                            endCursor
                            hasNextPage
                            hasPreviousPage
                            startCursor
                        }}
                    }}
                }}
            """)
            answer = "Here are some products I found:\n"
            products = json.loads(response)['data']['products']['nodes']
            card_list = []
            for product in products:
                product_dict = {
                    "id": product.get('id'),
                    "title": product.get('title'),
                    "description": product.get('description', "None")[:180] + "...",
                    "link_url": product.get('onlineStoreUrl'),
                    "image_url": product.get('images', {}).get('edges', [{}])[0].get('node', {}).get('src', ""), 
                    "variants": product.get('variants', {}).get('nodes', [])
                }
                card_list.append(product_dict)
            if card_list:
                try:
                    llm = PROVIDER_MAP.get(MODEL['llm_provider'], ChatOpenAI)(model=MODEL["model_type_or_path"], timeout=30000)
                    message = [
                        {"role": "system", "content": f"You are helping a customer search products based on the query and get results below and those results will be presented using product card format.\n\n{json.dumps(card_list)}\n\nGenerate a response to continue the conversation without explicitly mentioning contents of the search result. Include one or two questions about those products to know the user's preference. Keep the response within 50 words.\nDIRECTLY GIVE THE RESPONSE."},
                    ]
                    answer = llm.invoke(message).content
                except Exception as e:
                    logger.info(f"llm error in search_products: {e}")
                    pass
                return json.dumps({
                    "answer": answer,
                    "card_list": card_list
                })
            else:
                return json.dumps({
                    "answer": NO_PRODUCTS_FOUND_ERROR,
                    "card_list": []
                })
    
    except Exception as e:
        return json.dumps({
            "answer": NO_PRODUCTS_FOUND_ERROR,
            "card_list": []
        })


from benchmark.tau_bench.model_utils.api.api import API as API
from benchmark.tau_bench.model_utils.api.api import default_api_from_args as default_api_from_args
from benchmark.tau_bench.model_utils.api.api import BinaryClassifyDatapoint as BinaryClassifyDatapoint
from benchmark.tau_bench.model_utils.api.api import ClassifyDatapoint as ClassifyDatapoint
from benchmark.tau_bench.model_utils.api.api import GenerateDatapoint as GenerateDatapoint
from benchmark.tau_bench.model_utils.api.api import ParseDatapoint as ParseDatapoint
from benchmark.tau_bench.model_utils.api.api import ParseForceDatapoint as ParseForceDatapoint
from benchmark.tau_bench.model_utils.api.api import ScoreDatapoint as ScoreDatapoint
from benchmark.tau_bench.model_utils.api.api import default_api as default_api
from benchmark.tau_bench.model_utils.api.api import default_quick_api as default_quick_api
from benchmark.tau_bench.model_utils.api.datapoint import Datapoint as Datapoint
from benchmark.tau_bench.model_utils.api.datapoint import EvaluationResult as EvaluationResult
from benchmark.tau_bench.model_utils.api.datapoint import datapoint_factory as datapoint_factory
from benchmark.tau_bench.model_utils.api.datapoint import load_from_disk as load_from_disk
from benchmark.tau_bench.model_utils.api.exception import APIError as APIError
from benchmark.tau_bench.model_utils.api.sample import (
    EnsembleSamplingStrategy as EnsembleSamplingStrategy,
)
from benchmark.tau_bench.model_utils.api.sample import (
    MajoritySamplingStrategy as MajoritySamplingStrategy,
)
from benchmark.tau_bench.model_utils.api.sample import (
    RedundantSamplingStrategy as RedundantSamplingStrategy,
)
from benchmark.tau_bench.model_utils.api.sample import RetrySamplingStrategy as RetrySamplingStrategy
from benchmark.tau_bench.model_utils.api.sample import SamplingStrategy as SamplingStrategy
from benchmark.tau_bench.model_utils.api.sample import SingleSamplingStrategy as SingleSamplingStrategy
from benchmark.tau_bench.model_utils.api.sample import (
    UnanimousSamplingStrategy as UnanimousSamplingStrategy,
)
from benchmark.tau_bench.model_utils.api.sample import (
    get_default_sampling_strategy as get_default_sampling_strategy,
)
from benchmark.tau_bench.model_utils.api.sample import (
    set_default_sampling_strategy as set_default_sampling_strategy,
)
from benchmark.tau_bench.model_utils.model.chat import PromptSuffixStrategy as PromptSuffixStrategy
from benchmark.tau_bench.model_utils.model.exception import ModelError as ModelError
from benchmark.tau_bench.model_utils.model.general_model import GeneralModel as GeneralModel
from benchmark.tau_bench.model_utils.model.general_model import default_model as default_model
from benchmark.tau_bench.model_utils.model.general_model import model_factory as model_factory
from benchmark.tau_bench.model_utils.model.model import BinaryClassifyModel as BinaryClassifyModel
from benchmark.tau_bench.model_utils.model.model import ClassifyModel as ClassifyModel
from benchmark.tau_bench.model_utils.model.model import GenerateModel as GenerateModel
from benchmark.tau_bench.model_utils.model.model import ParseForceModel as ParseForceModel
from benchmark.tau_bench.model_utils.model.model import ParseModel as ParseModel
from benchmark.tau_bench.model_utils.model.model import Platform as Platform
from benchmark.tau_bench.model_utils.model.model import ScoreModel as ScoreModel
from benchmark.tau_bench.model_utils.model.openai import OpenAIModel as OpenAIModel
from benchmark.tau_bench.model_utils.model.utils import InputType as InputType



import logging

from langchain.prompts import PromptTemplate
from langchain_openai import ChatOpenAI
from langchain_core.output_parsers import StrOutputParser

from arklex.env.prompts import load_prompts
from arklex.types import EventType
from arklex.utils.utils import chunk_string
from arklex.utils.graph_state import MessageState
from arklex.utils.model_config import MODEL
from arklex.utils.model_provider_config import PROVIDER_MAP


logger = logging.getLogger(__name__)
    

class ToolGenerator():
    @staticmethod
    def generate(state: MessageState):
        user_message = state['user_message']
        
        prompts = load_prompts(state["bot_config"])
        llm = PROVIDER_MAP.get(MODEL['llm_provider'], ChatOpenAI)(
            model=MODEL["model_type_or_path"], timeout=30000, temperature=0.1
        )
        prompt = PromptTemplate.from_template(prompts["generator_prompt"])
        input_prompt = prompt.invoke({"sys_instruct": state["sys_instruct"], "formatted_chat": user_message.history})
        chunked_prompt = chunk_string(input_prompt.text, tokenizer=MODEL["tokenizer"], max_length=MODEL["context"])
        final_chain = llm | StrOutputParser()
        answer = final_chain.invoke(chunked_prompt)

        state["response"] = answer
        return state

    @staticmethod
    def context_generate(state: MessageState):
        llm = PROVIDER_MAP.get(MODEL['llm_provider'], ChatOpenAI)(
            model=MODEL["model_type_or_path"], timeout=30000, temperature=0.1
        )
        # get the input message
        user_message = state['user_message']
        message_flow = state['message_flow']
        logger.info(f"Retrieved texts (from retriever/search engine to generator): {message_flow[:50]} ...")
        
        # generate answer based on the retrieved texts
        prompts = load_prompts(state["bot_config"])
        prompt = PromptTemplate.from_template(prompts["context_generator_prompt"])
        input_prompt = prompt.invoke({"sys_instruct": state["sys_instruct"], "formatted_chat": user_message.history, "context": message_flow})
        chunked_prompt = chunk_string(input_prompt.text, tokenizer=MODEL["tokenizer"], max_length=MODEL["context"])
        final_chain = llm | StrOutputParser()
        logger.info(f"Prompt: {input_prompt.text}")
        answer = final_chain.invoke(chunked_prompt)
        state["message_flow"] = ""
        state["response"] = answer

        return state
    
    @staticmethod
    def stream_context_generate(state: MessageState):
        llm = PROVIDER_MAP.get(MODEL['llm_provider'], ChatOpenAI)(
            model=MODEL["model_type_or_path"], timeout=30000, temperature=0.1
        )
        # get the input message
        user_message = state['user_message']
        message_flow = state['message_flow']
        logger.info(f"Retrieved texts (from retriever/search engine to generator): {message_flow[:50]} ...")
        
        # generate answer based on the retrieved texts
        prompts = load_prompts(state["bot_config"])
        prompt = PromptTemplate.from_template(prompts["context_generator_prompt"])
        input_prompt = prompt.invoke({"sys_instruct": state["sys_instruct"], "formatted_chat": user_message.history, "context": message_flow})
        chunked_prompt = chunk_string(input_prompt.text, tokenizer=MODEL["tokenizer"], max_length=MODEL["context"])
        final_chain = llm | StrOutputParser()
        logger.info(f"Prompt: {input_prompt.text}")
        answer = ""
        for chunk in final_chain.stream(chunked_prompt):
            answer += chunk
            state["message_queue"].put({"event": EventType.CHUNK.value, "message_chunk": chunk})

        state["message_flow"] = ""
        state["response"] = answer

        return state
    
    @staticmethod
    def stream_generate(state: MessageState):
        user_message = state['user_message']
        
        prompts = load_prompts(state["bot_config"])
        llm = PROVIDER_MAP.get(MODEL['llm_provider'], ChatOpenAI)(
            model=MODEL["model_type_or_path"], timeout=30000, temperature=0.1
        )
        prompt = PromptTemplate.from_template(prompts["generator_prompt"])
        input_prompt = prompt.invoke({"sys_instruct": state["sys_instruct"], "formatted_chat": user_message.history})
        chunked_prompt = chunk_string(input_prompt.text, tokenizer=MODEL["tokenizer"], max_length=MODEL["context"])
        final_chain = llm | StrOutputParser()
        answer = ""
        for chunk in final_chain.stream(chunked_prompt):
            answer += chunk
            state["message_queue"].put({"event": EventType.CHUNK.value, "message_chunk": chunk})

        state["response"] = answer
        return state



import json
import networkx as nx

from arklex.evaluation.chatgpt_utils import chatgpt_chatbot, format_chat_history_str, flip_hist_content_only, filter_convo

def get_edges_and_counts(data):
    edge_counts = {}
    for convo in data:
        convo = filter_convo(convo)
        for i in range(len(convo)):
            if convo[i]['role'] == 'assistant':
                continue
            prev_intent = 'start' if i == 0 else convo[i-2]['intent']
            current_intent = convo[i]['intent']
            edge_counts[(prev_intent, current_intent)] = edge_counts.get((prev_intent, current_intent), 0) + 1
    return edge_counts

def build_intent_graph(data):
    G = nx.DiGraph()
    edge_counts = get_edges_and_counts(data)
    for key in edge_counts.keys():
        G.add_edge(key[0], key[1], weight = edge_counts[key])
    return G

def check_bot_goal(convo, bot_goal):
    convo_str = format_chat_history_str(flip_hist_content_only(convo[:-2]))
    prompt = f"Here is a conversation between a user and a customer service chatbot assistant:\n{convo_str}\n\nThe chatbot's goal is the following: {bot_goal}\nOutput True if the bot was able to achieve its goal. Output False otherwise. Only output True or False and nothing else."
    output = chatgpt_chatbot([{'role': 'user', 'content': prompt}])
    return output == "True"

def num_user_turns(convo):
    user_turns = 0
    for turn in convo:
        if turn.get('role', None) == 'user':
            user_turns += 1
    return user_turns

def extract_task_completion_metrics(data, bot_goal=None):
    num_convos = len(data)
    if num_convos == 0:
        return "Error while extracting task completion metrics"
    goal_completetions = 0
    bot_goal_completions = 0
    completion_efficiency = 0
    for convo in data:
        completion_efficiency += num_user_turns(convo)
        if convo[-1].get('goal_completetion', False):
            goal_completetions += 1
        if bot_goal is not None and check_bot_goal(convo, bot_goal):
            bot_goal_completions += 1
    metrics = {'user_task_completion': goal_completetions/num_convos,
               'user_task_completion_efficiency': completion_efficiency/num_convos}
    if bot_goal is not None:
        metrics['bot_goal_completion'] = bot_goal_completions/num_convos
    return metrics

if __name__ == "__main__":
    # with open('files/p1_sample_convos.json') as f:
    #     data = json.load(f)

    # model_api = "http://adaptation.cs.columbia.edu:55131/predict"
    # model_params = {'bot_id' : 'richtech', 'bot_version': 'v1alpha1'}
    # convos  = get_nlu_labels(data, model_api, model_params)
    # with open('files/p1_sample_convos_labeled.json', 'w') as f:
    #     json.dump(convos, f, indent=5)

    
    with open('files/p1_sample_convos_labeled.json') as f:
        data = json.load(f)
    G = build_intent_graph(data)
    for e in list(G.edges()):
        print(f"Weight for edge {e}: {G.get_edge_data(e[0], e[1])['weight']}")



# Copyright Sierra

from typing import Any, Dict
from benchmark.tau_bench.envs.tool import Tool


class Calculate(Tool):
    @staticmethod
    def invoke(data: Dict[str, Any], expression: str) -> str:
        if not all(char in "0123456789+-*/(). " for char in expression):
            return "Error: invalid characters in expression"
        try:
            # Evaluate the mathematical expression safely
            return str(round(float(eval(expression, {"__builtins__": None}, {})), 2))
        except Exception as e:
            return f"Error: {e}"

    @staticmethod
    def get_info() -> Dict[str, Any]:
        return {
            "type": "function",
            "function": {
                "name": "calculate",
                "description": "Calculate the result of a mathematical expression.",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "expression": {
                            "type": "string",
                            "description": "The mathematical expression to calculate, such as '2 + 2'. The expression can contain numbers, operators (+, -, *, /), parentheses, and spaces.",
                        },
                    },
                    "required": ["expression"],
                },
            },
        }



import os
import json
import argparse
import time
import logging
import subprocess
import signal
import atexit
from dotenv import load_dotenv
from pprint import pprint

import shopify

from arklex.utils.utils import init_logger
from arklex.orchestrator.orchestrator import AgentOrg
# from create import API_PORT
from arklex.utils.model_config import MODEL
from arklex.utils.model_provider_config import LLM_PROVIDERS
from arklex.env.env import Env

load_dotenv()
# session = shopify.Session(os.environ["SHOPIFY_SHOP_URL"], os.environ["SHOPIFY_API_VERSION"], os.environ["SHOPIFY_ACCESS_TOKEN"])
# shopify.ShopifyResource.activate_session(session)

process = None  # Global reference for the FastAPI subprocess

def pprint_with_color(data, color_code="\033[34m"):  # Default to blue
    print(color_code, end="")  # Set the color
    pprint(data)
    print("\033[0m", end="")  

def terminate_subprocess():
    """Terminate the FastAPI subprocess."""
    global process
    if process and process.poll() is None:  # Check if process is running
        logger.info(f"Terminating FastAPI process with PID: {process.pid}")
        process.terminate()  # Send SIGTERM
        process.wait()  # Ensure it stops
        logger.info("FastAPI process terminated.")

# Register cleanup function to run on program exit
atexit.register(terminate_subprocess)

# Handle signals (e.g., Ctrl+C)
signal.signal(signal.SIGINT, lambda signum, frame: exit(0))
signal.signal(signal.SIGTERM, lambda signum, frame: exit(0))


def get_api_bot_response(args, history, user_text, parameters, env):
    data = {"text": user_text, 'chat_history': history, 'parameters': parameters}
    orchestrator = AgentOrg(config=os.path.join(args.input_dir, "taskgraph.json"), env=env)
    result = orchestrator.get_response(data)

    return result['answer'], result['parameters']


def start_apis():
    """Start the FastAPI subprocess and update task graph API URLs."""
    global process
    command = [
        "uvicorn",
        "arklex.orchestrator.NLU.api:app",  # Replace with proper import path
        "--port", API_PORT,
        "--host", "0.0.0.0",
        "--log-level", "info"
    ]

    # Redirect FastAPI logs to a file
    with open("./logs/api.log", "w") as log_file:
        process = subprocess.Popen(
            command,
            stdout=log_file,  # Redirect stdout to a log file
            stderr=subprocess.STDOUT,  # Redirect stderr to the same file
            start_new_session=True  # Run in a separate process group
        )
    logger.info(f"Started FastAPI process with PID: {process.pid}")



if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('--input-dir', type=str, default="./examples/test")
    parser.add_argument('--model', type=str, default=MODEL["model_type_or_path"])
    parser.add_argument( '--llm-provider',type=str,default=MODEL["llm_provider"],choices=LLM_PROVIDERS)
    parser.add_argument('--log-level', type=str, default="WARNING", choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"])
    args = parser.parse_args()
    os.environ["DATA_DIR"] = args.input_dir
    MODEL["model_type_or_path"] = args.model
    MODEL["llm_provider"] = args.llm_provider
    log_level = getattr(logging, args.log_level.upper(), logging.WARNING)
    logger = init_logger(log_level=log_level, filename=os.path.join(os.path.dirname(__file__), "logs", "arklex.log"))

    # Initialize NLU and Slotfill APIs
    # start_apis()

    # Initialize env
    config = json.load(open(os.path.join(args.input_dir, "taskgraph.json")))
    env = Env(
        tools = config.get("tools", []),
        workers = config.get("workers", []),
        slotsfillapi = config["slotfillapi"]
    )
        
    history = []
    params = {}
    user_prefix = "user"
    worker_prefix = "assistant"
    for node in config['nodes']:
        if node[1].get("type", "") == 'start':
            start_message = node[1]['attribute']["value"]
            break
    history.append({"role": worker_prefix, "content": start_message})
    pprint_with_color(f"Bot: {start_message}")
    try:
        while True:
            user_text = input("You: ")
            if user_text.lower() == "quit":
                break
            start_time = time.time()
            output, params = get_api_bot_response(args, history, user_text, params, env)
            history.append({"role": user_prefix, "content": user_text})
            history.append({"role": worker_prefix, "content": output})
            print(f"getAPIBotResponse Time: {time.time() - start_time}")
            pprint_with_color(f"Bot: {output}")
    finally:
        terminate_subprocess()  # Ensure the subprocess is terminated



from abc import ABC
from abc import abstractmethod
import pickle

from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain_core.documents import Document

class Loader(ABC):
    def __init__(self):
        pass

    @staticmethod
    def save(filepath, data):
        with open(filepath, "wb") as f:
            pickle.dump(data, f)

    @abstractmethod
    def load(self, filepath):
        raise NotImplementedError

    @abstractmethod
    def chunk(self, document_objs):
        text_splitter = RecursiveCharacterTextSplitter.from_tiktoken_encoder(encoding_name="cl100k_base", chunk_size=200, chunk_overlap=40)
        docs = []
        langchain_docs = []
        for doc in document_objs:
            splitted_text = text_splitter.split_text(doc.content)
            for i, txt in enumerate(splitted_text):
                docs.append(doc)
                langchain_docs.append(Document(page_content=txt, metadata={"source": doc.title}))
        return langchain_docs


import json
import time
from typing import Any, Dict
import logging
import uuid
import os
from typing import List, Dict, Any, Tuple
import ast
import copy
from arklex.env.env import Env
import janus
from dotenv import load_dotenv
from difflib import SequenceMatcher

from langchain_core.runnables import RunnableLambda
import langsmith as ls
from openai import OpenAI
from langchain_openai import ChatOpenAI


from arklex.orchestrator.task_graph import TaskGraph
from arklex.env.tools.utils import ToolGenerator
from arklex.orchestrator.NLU.nlu import SlotFilling
from arklex.orchestrator.prompts import RESPOND_ACTION_NAME, RESPOND_ACTION_FIELD_NAME, REACT_INSTRUCTION
from arklex.types import EventType, StreamType
from arklex.utils.graph_state import ConvoMessage, OrchestratorMessage, MessageState, StatusEnum, BotConfig, Slot
from arklex.utils.utils import init_logger, truncate_string, format_chat_history, format_truncated_chat_history
from arklex.orchestrator.NLU.nlu import NLU
from arklex.utils.trace import TraceRunName
from arklex.utils.model_config import MODEL
from arklex.utils.model_provider_config import PROVIDER_MAP
from arklex.env.planner.function_calling import aimessage_to_dict


load_dotenv()
logger = logging.getLogger(__name__)


class AgentOrg:
    def __init__(self, config, env: Env, **kwargs):
        if isinstance(config, dict):
            self.product_kwargs = config
        else:
            self.product_kwargs = json.load(open(config))
        self.user_prefix = "user"
        self.worker_prefix = "assistant"
        self.environment_prefix = "tool"
        self.__eos_token = "\n"
        self.task_graph = TaskGraph("taskgraph", self.product_kwargs)
        self.env = env

    def generate_next_step(
        self, messages: List[Dict[str, Any]], text:str
    ) -> Tuple[Dict[str, Any], str, float]:
        llm = PROVIDER_MAP.get(MODEL['llm_provider'], ChatOpenAI)(
                    model=MODEL["model_type_or_path"],
                    temperature = 0.0,
                )
        if MODEL['llm_provider'] == 'gemini':
            messages = [
                ("system",str(messages[0]['content']),),
                ("human", ""),
            ]
        elif MODEL['llm_provider'] == 'anthropic':
            messages = [
            ("human",str(messages[0]['content']),),
        ]
        res = llm.invoke(messages)        
        message = aimessage_to_dict(res)
        action_str = message['content'].split("Action:")[-1].strip()
        try:
            action_parsed = json.loads(action_str)
        except json.JSONDecodeError:
            # this is a hack
            logger.warning(f"Failed to parse action: {action_str}, choose respond action")
            action_parsed = {
                "name": RESPOND_ACTION_NAME,
                "arguments": {RESPOND_ACTION_FIELD_NAME: action_str},
            }
        if action_parsed.get("name"):
            action = action_parsed["name"]
        else:
            logger.warning(f"Failed to parse action: {action_str}, choose respond action")
            action = RESPOND_ACTION_NAME
        # issues with getting response_cost using langchain, set to 0.0 for now
        return message, action, 0.0


    def get_response(self, inputs: dict, stream_type: StreamType = None, message_queue: janus.SyncQueue = None) -> Dict[str, Any]:
        text = inputs["text"]
        chat_history = inputs["chat_history"]
        params = inputs["parameters"]
        params["timing"] = {}
        chat_history_copy = copy.deepcopy(chat_history)
        chat_history_copy.append({"role": self.user_prefix, "content": text})
        chat_history_str = format_chat_history(chat_history_copy)
        dialog_states = params.get("dialog_states", {})
        if dialog_states:
            params["dialog_states"] = {tool: [Slot(**slot_data) for slot_data in slots] for tool, slots in dialog_states.items()}
        else:
            params["dialog_states"] = {}
        metadata = params.get("metadata", {})
        metadata["chat_id"] = metadata.get("chat_id", str(uuid.uuid4()))
        metadata["turn_id"] = metadata.get("turn_id", 0) + 1
        metadata["tool_response"] = {}
        params["metadata"] = metadata
        params["history"] = params.get("history", [])
        if not params["history"]:
            params["history"] = copy.deepcopy(chat_history_copy)
        else:
            params["history"].append(chat_history_copy[-2])
            params["history"].append(chat_history_copy[-1])


        ##### TaskGraph Chain
        taskgraph_inputs = {
            "text": text,
            "chat_history_str": chat_history_str,
            "parameters": params  ## TODO: different params for different components
        }
        dt = time.time()
        taskgraph_chain = RunnableLambda(self.task_graph.get_node) | RunnableLambda(self.task_graph.postprocess_node)
        node_info, params = taskgraph_chain.invoke(taskgraph_inputs)
        params["timing"]["taskgraph"] = time.time() - dt
        logger.info("=============node_info=============")
        logger.info(f"The first node info is : {node_info}") # {'name': 'MessageWorker', 'attribute': {'value': 'If you are interested, you can book a calendly meeting https://shorturl.at/crFLP with us. Or, you can tell me your phone number, email address, and name; our expert will reach out to you soon.', 'direct': False, 'slots': {"<name>": {<attributes>}}}}
        node_status = params.get("node_status", {})
        params["node_status"] = node_status

        with ls.trace(name=TraceRunName.TaskGraph, inputs={"taskgraph_inputs": taskgraph_inputs}) as rt:
            rt.end(
                outputs={
                    "metadata": params.get("metadata"),
                    "timing": params.get("timing", {}),
                    "curr_node": {
                        "id": params.get("curr_node"),
                        "name": node_info.get("name"),
                        "attribute": node_info.get("attribute")
                    },
                    "curr_global_intent": params.get("curr_pred_intent"),
                    "dialog_states": params.get("dialog_states"),
                    "node_status": params.get("node_status")}, 
                metadata={"chat_id": metadata.get("chat_id"), "turn_id": metadata.get("turn_id")}
            )

        # Direct response
        node_attribute = node_info["attribute"]
        if node_attribute["value"].strip():
            if node_attribute.get("direct_response"):                    
                return_response = {
                    "answer": node_attribute["value"],
                    "parameters": params
                }
                # Change the dialog_states from Class object to dict
                if params.get("dialog_states"):
                    params["dialog_states"] = {tool: [s.model_dump() for s in slots] for tool, slots in params["dialog_states"].items()}
                if node_attribute.get("type", "") == "multiple-choice" and node_attribute.get("choice_list", []):
                    return_response["choice_list"] = node_attribute["choice_list"]
                return return_response

        # Tool/Worker
        user_message = ConvoMessage(history=chat_history_str, message=text)
        orchestrator_message = OrchestratorMessage(message=node_info["attribute"]["value"], attribute=node_info["attribute"])
        sys_instruct = "You are a " + self.product_kwargs["role"] + ". " + self.product_kwargs["user_objective"] + self.product_kwargs["builder_objective"] + self.product_kwargs["intro"] + self.product_kwargs.get("opt_instruct", "")
        logger.info("=============sys_instruct=============")
        logger.info(sys_instruct)
        bot_config = BotConfig(
            bot_id=self.product_kwargs.get("bot_id", "default"),
            version=self.product_kwargs.get("version", "default"),
            language=self.product_kwargs.get("language", "EN"),
            bot_type=self.product_kwargs.get("bot_type", "presalebot"),
            available_workers=self.product_kwargs.get("workers", [])
        )
        message_state = MessageState(
            sys_instruct=sys_instruct, 
            bot_config=bot_config,
            user_message=user_message, 
            orchestrator_message=orchestrator_message, 
            trajectory=params["history"], 
            message_flow=params.get("worker_response", {}).get("message_flow", ""), 
            slots=params.get("dialog_states"),
            metadata=params.get("metadata"),
            is_stream=True if stream_type is not None else False,
            message_queue=message_queue
        )
        
        response_state, params = self.env.step(node_info["id"], message_state, params)
        
        logger.info(f"{response_state=}")

        tool_response = params.get("metadata", {}).get("tool_response", {})
        params["metadata"]["tool_response"] = {}

        with ls.trace(name=TraceRunName.ExecutionResult, inputs={"message_state": message_state}) as rt:
            rt.end(
                outputs={"metadata": params.get("metadata"), **response_state}, 
                metadata={"chat_id": metadata.get("chat_id"), "turn_id": metadata.get("turn_id")}
            )

        # ReAct framework to decide whether return to user or continue
        FINISH = False
        count = 0
        max_count = 5
        while not FINISH:
            # if the last response is from the assistant with content(which means not from tool or worker but from function calling response), 
            # then directly return the response otherwise it will continue to the next node but treat the previous response has been return to user.
            if response_state.get("trajectory", []) \
                and response_state["trajectory"][-1]["role"] == "assistant" \
                and response_state["trajectory"][-1]["content"]: 
                response_state["response"] = response_state["trajectory"][-1]["content"]
                break
            
            # If the max_count is reached, then break the loop
            if count >= max_count:
                logger.info("Max count reached, break the ReAct loop")
                break
            count += 1
            
            # If the current node is not complete, then no need to continue to the next node
            node_status = params.get("node_status", {})
            curr_node = params.get("curr_node", None)
            status = node_status.get(curr_node, StatusEnum.COMPLETE.value)
            if status == StatusEnum.INCOMPLETE.value:
                break

            node_info, params = taskgraph_chain.invoke(taskgraph_inputs)
            logger.info("=============node_info=============")
            logger.info(f"The while node info is : {node_info}")
            message_state["orchestrator_message"] = OrchestratorMessage(message=node_info["attribute"]["value"], attribute=node_info["attribute"])
            if node_info["id"] not in self.env.workers and node_info["id"] not in self.env.tools:
                message_state = MessageState(
                    sys_instruct=sys_instruct, 
                    user_message=user_message, 
                    orchestrator_message=orchestrator_message, 
                    trajectory=params["history"], 
                    message_flow=params.get("worker_response", {}).get("message_flow", ""), 
                    slots=params.get("dialog_states"),
                    metadata=params.get("metadata"),
                    is_stream=True if stream_type is not None else False,
                    message_queue=message_queue
                )
                
                action, response_state, msg_history = self.env.planner.execute(message_state, params["history"])
                params["history"] = msg_history
                if action == RESPOND_ACTION_NAME:
                    FINISH = True
                else:
                    tool_response = {}
            else:
                if node_info["id"] in self.env.tools:
                    node_actions = [{"name": self.env.id2name[node_info["id"]], "arguments": self.env.tools[node_info["id"]]["execute"]().info}]
                elif node_info["id"] in self.env.workers:
                    node_actions = [{"name": self.env.id2name[node_info["id"]], "description": self.env.workers[node_info["id"]]["execute"]().description}]
                
                # If the Default Worker enters the loop, it is the default node. It may call the RAG worker and no information in the context (tool response) can be used.
                if node_info["id"] == self.env.name2id["DefaultWorker"]:
                    logger.info("Skip the DefaultWorker in ReAct framework because it is the default node and may call the RAG worker (context cannot be used)")
                    action = RESPOND_ACTION_NAME
                    FINISH = True
                    break
                # If the Message Worker enters the loop, ReAct framework cannot make a good decision between the MessageWorker and the RESPOND action.
                elif node_info["id"] == self.env.name2id["MessageWorker"]:
                    logger.info("Skip ReAct framework because it is hard to distinguish between the MessageWorker and the RESPOND action")
                    message_state["response"] = "" # clear the response cache generated from the previous steps in the same turn
                    response_state, params = self.env.step(self.env.name2id["MessageWorker"], message_state, params)
                    FINISH = True
                    break
                action_spaces = node_actions
                response_msg = truncate_string(response_state.get("message_flow", "") or response_state.get("response", ""))
                action_spaces.append({"name": RESPOND_ACTION_NAME, "arguments": {RESPOND_ACTION_FIELD_NAME: response_msg}})
                truncated_params_history_str = format_truncated_chat_history(params["history"])
                prompt = REACT_INSTRUCTION.format(
                    conversation_record=truncated_params_history_str,
                    available_tools=json.dumps(action_spaces),
                    task=node_info["attribute"].get("task", "None"),
                )
                logger.info(f"React instruction: {prompt}")
                messages: List[Dict[str, Any]] = [
                    {"role": "system", "content": prompt}
                ]
                _, action, _ = self.generate_next_step(messages, text)
                action_name_list = [action_space["name"] for action_space in action_spaces]
                action_simi_score = [SequenceMatcher(None, action, action_name).ratio() for action_name in action_name_list]
                action = action_name_list[action_simi_score.index(max(action_simi_score))]
                logger.info("Predicted action: " + action)
                if action == RESPOND_ACTION_NAME:
                    FINISH = True
                else:
                    message_state["response"] = "" # clear the response cache generated from the previous steps in the same turn
                    response_state, params = self.env.step(self.env.name2id[action], message_state, params)
                    tool_response = params.get("metadata", {}).get("tool_response", {})
        if not response_state.get("response", ""):
            logger.info("No response from the ReAct framework, do context generation")
            tool_response = {}
            if stream_type is None:
                response_state = ToolGenerator.context_generate(response_state)
            else:
                response_state = ToolGenerator.stream_context_generate(response_state)

        response = response_state.get("response", "")
        params["metadata"]["tool_response"] = {}
        # TODO: params["metadata"]["worker"] is not serialization, make it empty for now
        if params.get("dialog_states"):
            params["dialog_states"] = {tool: [s.model_dump() for s in slots] for tool, slots in params["dialog_states"].items()}
        params["metadata"]["worker"] = {}
        params["tool_response"] = tool_response
        output = {
            "answer": response,
            "parameters": params
        }

        with ls.trace(name=TraceRunName.OrchestResponse) as rt:
            rt.end(
                outputs={"metadata": params.get("metadata"), **output},
                metadata={"chat_id": metadata.get("chat_id"), "turn_id": metadata.get("turn_id")}
            )

        return output



import json
from typing import Any, Dict

import shopify

from arklex.env.tools.tools import register_tool
from arklex.env.tools.shopify.utils import authorify_admin
from arklex.env.tools.shopify.utils_slots import ShopifySlots, ShopifyOutputs

description = "Get the status and details of an order."
slots = [
    ShopifySlots.ORDERS_ID,
    ShopifySlots.QUERY_LIMIT
]
outputs = [
    ShopifyOutputs.ORDERS_DETAILS
]
ORDERS_NOT_FOUND = "error: order not found"
errors = [
    ORDERS_NOT_FOUND
]

@register_tool(description, slots, outputs, lambda x: x not in errors)
def get_order_details(order_ids: list, limit=10, **kwargs) -> str:
    limit = int(limit) if limit else 10
    auth = authorify_admin(kwargs)
    if auth["error"]:
        return auth["error"]
    
    try:
        ids = ' OR '.join(f'id:{oid.split("/")[-1]}' for oid in order_ids)
        with shopify.Session.temp(**auth["value"]):
            response = shopify.GraphQL().execute(f"""
            {{
                orders (first: {limit}, query:"{ids}") {{
                    nodes {{
                        id
                        name
                        createdAt
                        cancelledAt
                        returnStatus
                        statusPageUrl
                        totalPriceSet {{
                            presentmentMoney {{
                                amount
                            }}
                        }}
                        fulfillments {{
                            displayStatus
                            trackingInfo {{
                                number
                                url
                            }}
                        }}
                        lineItems(first: 10) {{
                            edges {{
                                node {{
                                    id
                                    title
                                    quantity
                                    variant {{
                                        id
                                        product {{
                                            id
                                        }}
                                    }}
                                }}
                            }}
                        }}
                    }}
                }}
            }}
            """)
            result = json.loads(response)["data"]["orders"]["nodes"]
            response_text = ""
            for order in result:
                response_text += f"Order ID: {order.get('id', 'None')}\n"
                response_text += f"Order Name: {order.get('name', 'None')}\n"
                response_text += f"Created At: {order.get('createdAt', 'None')}\n"
                response_text += f"Cancelled At: {order.get('cancelledAt', 'None')}\n"
                response_text += f"Return Status: {order.get('returnStatus', 'None')}\n"
                response_text += f"Status Page URL: {order.get('statusPageUrl', 'None')}\n"
                response_text += f"Total Price: {order.get('totalPriceSet', {}).get('presentmentMoney', {}).get('amount', 'None')}\n"
                response_text += f"Fulfillment Status: {order.get('fulfillments', 'None')}\n"
                response_text += "Line Items:\n"
                for item in order.get('lineItems', {}).get('edges', []):
                    response_text += f"    Title: {item.get('node', {}).get('title', 'None')}\n"
                    response_text += f"    Quantity: {item.get('node', {}).get('quantity', 'None')}\n"
                    response_text += f"    Variant ID: {item.get('node', {}).get('variant', {}).get('id', 'None')}\n"
                response_text += "\n"
        return response_text
    except Exception as e:
        return ORDERS_NOT_FOUND



# Copyright Sierra

from typing import Any, Dict
from benchmark.tau_bench.envs.tool import Tool


class FindUserIdByEmail(Tool):
    @staticmethod
    def invoke(data: Dict[str, Any], email: str) -> str:
        users = data["users"]
        for user_id, profile in users.items():
            if profile["email"].lower() == email.lower():
                return user_id
        return "Error: user not found"

    @staticmethod
    def get_info() -> Dict[str, Any]:
        return {
            "type": "function",
            "function": {
                "name": "find_user_id_by_email",
                "description": "Find user id by email. If the user is not found, the function will return an error message.",
                "parameters": {
                    "type": "object",
                    "properties": {
                        "email": {
                            "type": "string",
                            "description": "The email of the user, such as 'something@example.com'.",
                        },
                    },
                    "required": ["email"],
                },
            },
        }



# Copyright Sierra

import os
import json
import random
import traceback
from math import comb
import multiprocessing
from typing import List, Dict, Any
from datetime import datetime
from concurrent.futures import ThreadPoolExecutor

from benchmark.tau_bench.envs import get_env
from benchmark.tau_bench.agents.base import Agent
from benchmark.tau_bench.types import EnvRunResult, RunConfig
from litellm import provider_list
from benchmark.tau_bench.envs.user import UserStrategy


def run(config: RunConfig) -> List[EnvRunResult]:
    assert config.env in ["retail", "airline"], "Only retail and airline envs are supported"
    assert config.user_model_provider in provider_list, "Invalid user model provider"
    assert config.task_split in ["train", "test", "dev"], "Invalid task split"
    assert config.user_strategy in [item.value for item in UserStrategy], "Invalid user strategy"

    random.seed(config.seed)
    ckpt_path = f"{config.output_dir}/tau_bench_evaluation.json"
    if not os.path.exists(config.log_dir):
        os.makedirs(config.log_dir)

    print(f"Loading user with strategy: {config.user_strategy}")
    env = get_env(
        config.env,
        user_strategy=config.user_strategy,
        user_model=config.user_model,
        user_provider=config.user_model_provider,
        task_split=config.task_split,
    )
    agent = agent_factory(config=config)
    end_index = (
        len(env.tasks) if config.end_index == -1 else min(config.end_index, len(env.tasks))
    )
    results: List[EnvRunResult] = []
    lock = multiprocessing.Lock()
    if config.task_ids and len(config.task_ids) > 0:
        print(f"Running tasks {config.task_ids} (checkpoint path: {ckpt_path})")
    else:
        print(
            f"Running tasks {config.start_index} to {end_index} (checkpoint path: {ckpt_path})"
    )
    for i in range(config.num_trials):
        if config.task_ids and len(config.task_ids) > 0:
            idxs = config.task_ids
        else:
            idxs = list(range(config.start_index, end_index))
        if config.shuffle:
            random.shuffle(idxs)

        def _run(idx: int) -> EnvRunResult:
            isolated_env = get_env(
                config.env,
                user_strategy=config.user_strategy,
                user_model=config.user_model,
                task_split=config.task_split,
                user_provider=config.user_model_provider,
                task_index=idx,
            )

            print(f"Running task {idx}")
            try:
                res = agent.solve(
                    env=isolated_env,
                    task_index=idx,
                )
                result = EnvRunResult(
                    task_id=idx,
                    reward=res.reward,
                    info=res.info,
                    traj=res.messages,
                    trial=i,
                )
            except Exception as e:
                result = EnvRunResult(
                    task_id=idx,
                    reward=0.0,
                    info={"error": str(e), "traceback": traceback.format_exc()},
                    traj=[],
                    trial=i,
                )
            print(
                "✅" if result.reward == 1 else "❌",
                f"task_id={idx}",
                result.info,
            )
            print("-----")
            with lock:
                data = []
                with open(ckpt_path, "w") as f:
                    json.dump(data + [result.model_dump()], f, indent=2)
            return result

        with ThreadPoolExecutor(max_workers=config.max_concurrency) as executor:
            res = list(executor.map(_run, idxs))
            results.extend(res)

    avg_reward, pass_hat_ks = get_metrics(results)
    display_metrics(avg_reward, pass_hat_ks)

    tasks_res = [result.model_dump() for result in results]
    with open(ckpt_path, "w") as f:
        tau_bench_evaluation = {
            "average_reward": avg_reward,
            "pass_k": pass_hat_ks,
            "task_results": tasks_res
        }
        json.dump(tau_bench_evaluation, f, indent=2)
        print(f"\n📄 Results saved to {ckpt_path}\n")
    return results


def agent_factory(config: RunConfig) -> Agent:
    from benchmark.tau_bench.agents.agent_first_org import AgentFirstOrg
    return AgentFirstOrg(taskgraph_dir=config.taskgraph_dir)


def get_metrics(results: List[EnvRunResult]):
    def is_successful(reward: float) -> bool:
        return (1 - 1e-6) <= reward <= (1 + 1e-6)

    num_trials = len(set([r.trial for r in results]))
    rewards = [r.reward for r in results]
    avg_reward = sum(rewards) / len(rewards)
    # c from https://arxiv.org/pdf/2406.12045
    c_per_task_id: dict[int, int] = {}
    for result in results:
        if result.task_id not in c_per_task_id:
            c_per_task_id[result.task_id] = 1 if is_successful(result.reward) else 0
        else:
            c_per_task_id[result.task_id] += 1 if is_successful(result.reward) else 0
    pass_hat_ks: dict[int, float] = {}
    for k in range(1, num_trials + 1):
        sum_task_pass_hat_k = 0
        for c in c_per_task_id.values():
            sum_task_pass_hat_k += comb(c, k) / comb(num_trials, k)
        pass_hat_ks[k] = sum_task_pass_hat_k / len(c_per_task_id)
    return avg_reward, pass_hat_ks

def display_metrics(avg_reward, pass_hat_ks) -> None:
    print(f"🏆 Average reward: {avg_reward}")
    print("📈 Pass^k")
    for k, pass_hat_k in pass_hat_ks.items():
        print(f"  k={k}: {pass_hat_k}")
    



import json

import shopify

from arklex.utils.loaders.base import Loader
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain_core.documents import Document

class ShopifyLoader(Loader):
    def __init__(self):
        pass

    def load(self):
        docs = []
        response = shopify.GraphQL().execute("""
            {
                products(first: 23) {
                    edges {
                        node {
                            title
                            tags
                            description
                            totalInventory
                        }
                    }
                }
            }
            """)
        product_docs = json.loads(response)["data"]["products"]["edges"]
        for product_doc in product_docs:
            docs.append(Document(page_content=product_doc["node"]["description"], metadata={"title": product_doc["node"]["title"]}))
        return docs

    def chunk(self, document_objs):
        text_splitter = RecursiveCharacterTextSplitter.from_tiktoken_encoder(encoding_name="cl100k_base", chunk_size=200, chunk_overlap=40)
        docs = []
        langchain_docs = []
        for doc in document_objs:
            splitted_text = text_splitter.split_text(doc.page_content)
            for i, txt in enumerate(splitted_text):
                docs.append(doc)
                langchain_docs.append(Document(page_content=txt, metadata=doc.metadata))
        return langchain_docs


import requests
import logging
from dotenv import load_dotenv

import langsmith as ls

from arklex.utils.model_config import MODEL
from arklex.utils.trace import TraceRunName
from arklex.utils.graph_state import Slots, Slot
from arklex.orchestrator.NLU.api import nlu_api, slotfilling_api

load_dotenv()
logger = logging.getLogger(__name__)


class NLU:
    def __init__(self, url):
        self.url = url

    def execute(self, text:str, intents:dict, chat_history_str:str, metadata:dict) -> str:
        logger.info(f"candidates intents of NLU: {intents}")
        data = {
            "text": text,
            "intents": intents,
            "chat_history_str": chat_history_str,
            "model":MODEL
        }
        if self.url:
            logger.info(f"Using NLU API to predict the intent")
            response = requests.post(self.url + "/predict", json=data)
            if response.status_code == 200:
                results = response.json()
                pred_intent = results['intent']
                logger.info(f"pred_intent is {pred_intent}")
            else:
                pred_intent = "others"
                logger.error('Remote Server Error when predicting NLU')
        else:
            logger.info(f"Using NLU function to predict the intent")
            pred_intent = nlu_api.predict(**data)
            logger.info(f"pred_intent is {pred_intent}")

        with ls.trace(name=TraceRunName.NLU, inputs=data) as rt:
            rt.end(
                outputs={"pred_intent": pred_intent},
                metadata={"chat_id": metadata.get("chat_id"), "turn_id": metadata.get("turn_id")}
            )
        return pred_intent
    

class SlotFilling:
    def __init__(self, url):
        self.url = url

    def verify_needed(self, slot: Slot, chat_history_str:str, metadata: dict) -> Slot:
        logger.info(f"verify slot: {slot}")
        data = {
            "slot": slot.model_dump(),
            "chat_history_str": chat_history_str,
            "model":MODEL
        }
        if self.url:
            logger.info(f"Using Slot Filling API to verify the slot")
            response = requests.post(self.url + "/verify", json=data)
            if response.status_code == 200:
                verification_needed = response.json().get("verification_needed")
                thought = response.json().get("thought")
                logger.info(f"verify_needed is {verification_needed}")
            else:
                verification_needed = False
                thought = "No need to verify"
                logger.error('Remote Server Error when verifying Slot Filling')
        else:
            logger.info(f"Using Slot Filling function to verify the slot")
            verification = slotfilling_api.verify(**data)
            verification_needed = verification.verification_needed
            thought = verification.thought
            logger.info(f"verify_needed is {verification_needed}")

        return verification_needed, thought

    def execute(self, slots:list, chat_history_str:str, metadata: dict) -> dict:
        logger.info(f"extracted slots: {slots}")
        if not slots: return []
        
        data = {
            "slots": [slot.model_dump() for slot in slots],
            "chat_history_str": chat_history_str,
            "model":MODEL
        }
        if self.url:
            logger.info(f"Using Slot Filling API to predict the slots")
            response = requests.post(self.url + "/predict", json=data)
            if response.status_code == 200:
                pred_slots = response.json()
                pred_slots = [Slot(**pred_slot) for pred_slot in pred_slots]
                logger.info(f"pred_slots is {pred_slots}")
            else:
                pred_slots = slots
                logger.error('Remote Server Error when predicting Slot Filling')
        else:
            logger.info(f"Using Slot Filling function to predict the slots")
            pred_slots = slotfilling_api.predict(**data).slots
            # pred_slots = [Slot(**pred_slot) for pred_slot in pred_slots]
            logger.info(f"pred_slots is {pred_slots}")
        with ls.trace(name=TraceRunName.SlotFilling, inputs=data) as rt:
            rt.end(
                outputs=pred_slots,
                metadata={"chat_id": metadata.get("chat_id"), "turn_id": metadata.get("turn_id")}
            )
        return pred_slots



import os
import sys
import json
import logging
from logging.handlers import RotatingFileHandler
from arklex.utils.model_config import MODEL

import tiktoken
import Levenshtein

logger = logging.getLogger(__name__)


def init_logger(log_level=logging.INFO, filename=None):
    root_logger = logging.getLogger()  # Root logger

    # Remove existing handlers to reconfigure them
    if root_logger.hasHandlers():
        root_logger.handlers.clear()
    handlers = []
    # File handler
    if filename is not None:
        directory_name, _ = os.path.split(filename)
        if not os.path.exists(directory_name):
            os.makedirs(directory_name)
        file_handler = RotatingFileHandler(
            filename=filename, 
            mode='a',
            maxBytes=50*1024*1024,
            backupCount=20,
            encoding=None,
            delay=0
        )
        file_handler.setLevel(log_level)  # Set log level for the file
        file_handler.setFormatter(logging.Formatter(
            "[%(asctime)s] {%(filename)s:%(lineno)d} %(levelname)s - %(message)s",
            datefmt="%m/%d/%Y %H:%M:%S"
        ))
        handlers.append(file_handler)

    # Stream (terminal) handler
    stream_handler = logging.StreamHandler(sys.stdout)
    stream_handler.setLevel(log_level)  # Set log level for the terminal
    stream_handler.setFormatter(logging.Formatter(
        "[%(asctime)s] {%(filename)s:%(lineno)d} %(levelname)s - %(message)s",
        datefmt="%m/%d/%Y %H:%M:%S"
    ))
    handlers.append(stream_handler)

    for handler in handlers:
        root_logger.addHandler(handler)
    root_logger.setLevel(log_level)

    # Suppress noisy loggers
    logging.getLogger("LiteLLM").setLevel(logging.ERROR)

    return logging.getLogger(__name__)


def chunk_string(text, tokenizer, max_length, from_end=True):
    # Initialize the tokenizer
	encoding = tiktoken.get_encoding(tokenizer)
	tokens = encoding.encode(text)
	if from_end:
		chunks = encoding.decode(tokens[-max_length:])
	else:
		chunks = encoding.decode(tokens[:max_length])
	return chunks

def normalize(lst):
		return [float(num)/sum(lst) for num in lst]

def str_similarity(string1, string2):
	try:
		distance = Levenshtein.distance(string1, string2)
		max_length = max(len(string1), len(string2))
		similarity = 1 - (distance / max_length)
	except Exception as err:
		print(err)
		similarity = 0
	return similarity


def postprocess_json(raw_code):
	valid_phrases = ['"', '{', '}', '[', ']']

	valid_lines = []
	for line in raw_code.split('\n'):
		if len(line) == 0:
			continue
		# If the line not starts with any of the valid phrases, skip it
		should_skip = not any([line.strip().startswith(phrase) for phrase in valid_phrases])
		if should_skip:
			continue
		valid_lines.append(line)

	try:
		generated_result = "\n".join(valid_lines)
		result = json.loads(generated_result)
	except json.JSONDecodeError as e:
		logger.error(f"Error decoding generated JSON - {generated_result}")
		logger.error(f"raw result: {raw_code}")
		logger.error(f"Error: {e}")
		result = None
	return result

def truncate_string(text: str, max_length: int=400):
    if len(text) > max_length:
        text = text[:max_length] + "..."
    return text

def format_chat_history(chat_history):
    '''Includes current user utterance'''
    chat_history_str= ""
    for turn in chat_history:
        chat_history_str += f"{turn['role']}: {turn['content']}\n"
    return chat_history_str.strip()

def format_truncated_chat_history(chat_history, max_length=400):
    '''Includes current user utterance'''
    chat_history_str= ""
    for turn in chat_history:
        chat_history_str += f"{turn['role']}: {truncate_string(turn['content'], max_length) if turn['content'] else turn['content']}\n"
    return chat_history_str.strip()

def format_messages_by_provider(messages, text, model=MODEL):
    llm_provider = model['llm_provider']
    if llm_provider == "anthropic":
        return {"system": f"{messages[0]['content']}"}
    elif llm_provider == 'gemini':
        messages.append({"role": "user", "content": f"{text}"})
    return {"messages": messages}
   


import os
import logging
from typing import List
import pickle

from langchain.prompts import PromptTemplate
from langchain_openai import ChatOpenAI
from langchain_core.output_parsers import StrOutputParser
from langchain_core.documents import Document
from langchain_community.vectorstores.faiss import FAISS
from langchain_openai import OpenAIEmbeddings


from arklex.utils.model_config import MODEL
from arklex.env.prompts import load_prompts
from arklex.utils.graph_state import MessageState
from arklex.utils.model_provider_config import PROVIDER_MAP, PROVIDER_EMBEDDINGS, PROVIDER_EMBEDDING_MODELS


logger = logging.getLogger(__name__)

class RetrieveEngine():
    @staticmethod
    def faiss_retrieve(state: MessageState):
        # get the input message
        user_message = state['user_message']

        # Search for the relevant documents
        prompts = load_prompts(state["bot_config"])
        docs = FaissRetrieverExecutor.load_docs(database_path=os.environ.get("DATA_DIR"))
        retrieved_text = docs.search(user_message.history, prompts["retrieve_contextualize_q_prompt"])

        state["message_flow"] = retrieved_text
        return state


class FaissRetrieverExecutor:
    def __init__(
            self, 
            texts: List[Document], 
            index_path: str,
            embedding_model_name: str =  PROVIDER_EMBEDDING_MODELS[MODEL['llm_provider']]
        ):
        self.texts = texts
        self.index_path = index_path
        self.embedding_model_name = embedding_model_name
        self.llm = PROVIDER_MAP.get(MODEL['llm_provider'], ChatOpenAI)(
            model=MODEL["model_type_or_path"], timeout=30000
        )
        self.retriever = self._init_retriever()

    def _init_retriever(self, **kwargs):
        # initiate FAISS retriever
        embedding_model = PROVIDER_EMBEDDINGS.get(MODEL['llm_provider'], OpenAIEmbeddings)(
            **{ 'model': self.embedding_model_name } if MODEL['llm_provider'] != 'anthropic' else { 'model_name': self.embedding_model_name }
        )
        docsearch = FAISS.from_documents(self.texts, embedding_model)
        retriever = docsearch.as_retriever(**kwargs)
        return retriever     

    def retrieve_w_score(self, query: str):
        k_value = 4 if not self.retriever.search_kwargs.get('k') else self.retriever.search_kwargs.get('k')
        docs_and_scores = self.retriever.vectorstore.similarity_search_with_score(query, k=k_value)
        return docs_and_scores

    def search(self, chat_history_str: str, contextualize_prompt: str):
        contextualize_q_prompt = PromptTemplate.from_template(
            contextualize_prompt
        )
        ret_input_chain = contextualize_q_prompt | self.llm | StrOutputParser()
        ret_input = ret_input_chain.invoke({"chat_history": chat_history_str})
        logger.info(f"Reformulated input for retriever search: {ret_input}")
        docs_and_score = self.retrieve_w_score(ret_input)
        retrieved_text = ""
        for doc, score in docs_and_score:
            retrieved_text += f"{doc.page_content} \n"
        return retrieved_text

    @staticmethod
    def load_docs(database_path: str, embeddings: str=None, index_path: str="./index"):
        document_path = os.path.join(database_path, "chunked_documents.pkl")
        index_path = os.path.join(database_path, "index")
        logger.info(f"Loaded documents from {document_path}")
        with open(document_path, 'rb') as fread:
            documents = pickle.load(fread)
        logger.info(f"Loaded {len(documents)} documents")

        return FaissRetrieverExecutor(
            texts=documents,
            index_path=index_path
        )



import json
from typing import Any, Dict
import shopify

# general GraphQL navigation utilities
from arklex.env.tools.shopify.utils_slots import ShopifySlots, ShopifyOutputs
from arklex.env.tools.shopify.utils_nav import *
from arklex.env.tools.shopify.utils import authorify_admin

from arklex.env.tools.tools import register_tool

description = "Get the details of collection with a preview of the inventory from a list of collections."
slots = [
    ShopifySlots.COLLECTION_IDS,
    *PAGEINFO_SLOTS
]
outputs = [
    ShopifyOutputs.COLLECTIONS_DETAILS,
    *PAGEINFO_OUTPUTS
]
COLLECTION_NOT_FOUND = "error: collection not found"
errors = [COLLECTION_NOT_FOUND]

@register_tool(description, slots, outputs, lambda x: x not in errors)
def preview_collections(collection_ids: list, **kwargs) -> str:
    nav = cursorify(kwargs)
    if not nav[1]:
        return nav[0]
    auth = authorify_admin(kwargs)
    if auth["error"]:
        return auth["error"]
    
    try:
        with shopify.Session.temp(**auth["value"]):
            ids = ' OR '.join(f'id:{cid.split("/")[-1]}' for cid in collection_ids)
            response = shopify.GraphQL().execute(f"""
                {{
                    collections ({nav[0]}, query:"{ids}") {{
                        nodes {{
                            title
                            description
                            productsCount {{
                                count
                            }}
                            products (first: 5) {{
                                nodes {{
                                    title
                                    description
                                    id
                                }}
                            }}
                        }}
                        pageInfo {{
                            endCursor
                            hasNextPage
                            hasPreviousPage
                            startCursor
                        }}
                    }}
                }}
            """)
            results = json.loads(response)['data']
            response = results['collections']["nodes"]
            pageInfo = results['collections']['pageInfo']
            return response, pageInfo
    except Exception as e:
        return COLLECTION_NOT_FOUND


from __future__ import annotations

import abc
import json
from typing import Any, Callable, TypeVar

from pydantic import BaseModel
from benchmark.tau_bench.model_utils.api._model_methods import MODEL_METHODS
from benchmark.tau_bench.model_utils.api.exception import APIError
from benchmark.tau_bench.model_utils.api.types import PartialObj
from benchmark.tau_bench.model_utils.model.exception import ModelError

T = TypeVar("T", bound=BaseModel)


def _is_trace(obj: dict[str, Any]) -> bool:
    return (
        "method_name" in obj
        and obj["method_name"] in MODEL_METHODS
        and "kwargs" in obj
        and "response" in obj
        and isinstance(obj["kwargs"], dict)
    )


def dict_equal(d1: dict, d2: dict) -> bool:
    d1_keys_sorted = sorted(d1.keys())
    d2_keys_sorted = sorted(d2.keys())
    if d1_keys_sorted != d2_keys_sorted:
        return False
    for k in d1_keys_sorted:
        if isinstance(d1[k], dict) and isinstance(d2[k], dict):
            if not dict_equal(d1[k], d2[k]):
                return False
        elif isinstance(d1[k], list) and isinstance(d2[k], list):
            if not list_equal(d1[k], d2[k]):
                return False
        elif isinstance(d1[k], set) and isinstance(d2[k], set):
            if d1[k] != d2[k]:
                return False
        elif isinstance(d1[k], str) and isinstance(d2[k], str):
            if not str_equal(d1[k], d2[k]):
                return False
        elif d1[k] != d2[k]:
            return False
    return True


def list_equal(l1: list, l2: list) -> bool:
    if len(l1) != len(l2):
        return False
    for i1, i2 in zip(l1, l2):
        if isinstance(i1, dict) and isinstance(i2, dict):
            if not dict_equal(i1, i2):
                return False
        elif isinstance(i1, list) and isinstance(i2, list):
            if not list_equal(i1, i2):
                return False
        elif isinstance(i1, set) and isinstance(i2, set):
            if i1 != i2:
                return False
        elif isinstance(i1, str) and isinstance(i2, str):
            if not str_equal(i1, i2):
                return False
        elif i1 != i2:
            return False
    return True


def set_equal(s1: set, s2: set) -> bool:
    if len(s1) != len(s2):
        return False
    for i1, i2 in zip(s1, s2):
        if isinstance(i1, dict) and isinstance(i2, dict):
            if not dict_equal(i1, i2):
                return False
        elif isinstance(i1, list) and isinstance(i2, list):
            if not list_equal(i1, i2):
                return False
        elif isinstance(i1, set) and isinstance(i2, set):
            if i1 != i2:
                return False
        elif isinstance(i1, str) and isinstance(i2, str):
            if not str_equal(i1, i2):
                return False
        elif i1 != i2:
            return False
    return True


def str_equal(s1: str, s2: str) -> bool:
    def remove_special_chars(s: str) -> str:
        return "".join(filter(str.isalnum, s))

    def strip_and_lower(s: str) -> str:
        return s.lower().strip()

    return strip_and_lower(remove_special_chars(s1)) == strip_and_lower(remove_special_chars(s2))


class EvaluationResult(BaseModel):
    is_error: bool
    is_correct: bool
    datapoint: dict[str, Any] | None
    response: Any | None
    error: str | None


class Datapoint(BaseModel, abc.ABC):
    @classmethod
    def from_trace(cls, d: dict[str, Any]) -> "Datapoint":
        if not _is_trace(d):
            raise ValueError(f"This is not a trace: {d}")
        response = d["response"]
        kwargs = d["kwargs"]
        return cls(response=response, **kwargs)

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> "Datapoint":
        if _is_trace(d):
            return cls.from_trace(d)
        return cls(**d)

    @abc.abstractmethod
    def evaluate(self, api: tau_bench.model_utils.API) -> EvaluationResult:
        raise NotImplementedError


class ClassifyDatapoint(Datapoint):
    instruction: str
    text: str
    options: list[str]
    response: int | None = None
    examples: list["ClassifyDatapoint"] | None = None

    def evaluate(self, api: tau_bench.model_utils.API) -> EvaluationResult:
        return run_and_catch_api_error(
            lambda: api.classify(
                instruction=self.instruction,
                text=self.text,
                options=self.options,
                examples=self.examples,
            ),
            self.response,
            self.model_dump(),
        )


class BinaryClassifyDatapoint(Datapoint):
    instruction: str
    text: str
    response: bool | None = None
    examples: list["BinaryClassifyDatapoint"] | None = None

    def evaluate(self, api: tau_bench.model_utils.API) -> EvaluationResult:
        return run_and_catch_api_error(
            lambda: api.binary_classify(
                instruction=self.instruction, text=self.text, examples=self.examples
            ),
            self.response,
            self.model_dump(),
        )


class ScoreDatapoint(Datapoint):
    instruction: str
    text: str
    min: int
    max: int
    response: int | None = None
    examples: list["ScoreDatapoint"] | None = None

    def evaluate(self, api: tau_bench.model_utils.API) -> EvaluationResult:
        raise NotImplementedError


class ParseDatapoint(Datapoint):
    text: str
    typ: type[T] | dict[str, Any]
    response: dict[str, Any] | T | PartialObj | None = None
    examples: list["ParseDatapoint"] | None = None

    def evaluate(self, api: tau_bench.model_utils.API) -> EvaluationResult:
        return run_and_catch_api_error(
            lambda: api.parse(text=self.text, typ=self.typ),
            self.response,
            self.model_dump(),
        )


class GenerateDatapoint(Datapoint):
    instruction: str
    text: str
    response: str | None = None
    examples: list["GenerateDatapoint"] | None = None

    def evaluate(self, api: tau_bench.model_utils.API) -> tau_bench.model_utils.EvaluationResult:
        raise NotImplementedError


class ParseForceDatapoint(Datapoint):
    instruction: str
    typ: type[T] | dict[str, Any]
    text: str | None = None
    response: dict[str, Any] | T | None = None
    examples: list["ParseForceDatapoint"] | None = None

    def evaluate(self, api: tau_bench.model_utils.API) -> EvaluationResult:
        return run_and_catch_api_error(
            lambda: api.parse_force(
                instruction=self.instruction,
                text=self.text,
                typ=self.typ,
                examples=self.examples,
            ),
            self.response,
            self.model_dump(),
        )


def datapoint_factory(d: dict[str, Any]) -> Datapoint:
    if _is_trace(d):
        method_name = d["method_name"]
        kwargs = d["kwargs"]
        data = {"response": d["response"], **kwargs}
        if method_name == "classify":
            return ClassifyDatapoint(**data)
        elif method_name == "binary_classify":
            return BinaryClassifyDatapoint(**data)
        elif method_name == "parse":
            return ParseDatapoint(**data)
        elif method_name == "parse_force":
            return ParseForceDatapoint(**data)
        elif method_name == "generate":
            return GenerateDatapoint(**data)
        elif method_name == "score":
            return ScoreDatapoint(**data)
        else:
            raise ValueError(f"Unknown method name: {method_name}")
    else:
        if all(k in d for k in ["instruction", "text", "options"]) and isinstance(
            d["response"], int
        ):
            return ClassifyDatapoint(**d)
        elif all(k in d for k in ["instruction", "text"]) and isinstance(d["response"], bool):
            return BinaryClassifyDatapoint(**d)
        elif all(k in d for k in ["instruction", "text", "min", "max"]) and isinstance(
            d["response"], int
        ):
            return ScoreDatapoint(**d)
        elif all(k in d for k in ["instruction", "text", "typ"]) and isinstance(
            d["response"], dict
        ):
            return ParseForceDatapoint(**d)
        elif all(k in d for k in ["text", "typ"]) and isinstance(d["response"], dict):
            return ParseDatapoint(**d)
        elif all(k in d for k in ["instruction", "text"]) and isinstance(d["response"], str):
            return GenerateDatapoint(**d)
        else:
            raise ValueError(f"Unknown datapoint: {d}")


def run_and_catch_api_error(
    callable: Callable[..., Any], response: Any, datapoint: dict[str, Any]
) -> EvaluationResult:
    try:
        res = callable()
        if isinstance(response, dict):
            is_correct = dict_equal(res, response)
        else:
            is_correct = res == response
        return EvaluationResult(
            is_error=False,
            is_correct=is_correct,
            response=res,
            error=None,
            datapoint=datapoint,
        )
    except (APIError, ModelError) as e:
        return EvaluationResult(
            is_error=True,
            is_correct=False,
            response=None,
            error=str(e),
            datapoint=datapoint,
        )


def load_from_disk(path: str) -> list[Datapoint]:
    with open(path, "r") as f:
        if path.endswith(".jsonl"):
            data = [json.loads(line) for line in f]
        elif path.endswith(".json"):
            data = json.load(f)
        else:
            raise ValueError(f"Unknown file format: {path}")
    return [datapoint_factory(d) for d in data]
